@@ -12,7 +12,6 @@
 #include "lib/resource.h"
 #include "lib/timer.h"
 
-#define BIRD_FNAME_MAX 255	/* Would be better to use some UNIX define */
 
 /* Configuration structure */
 
@@ -27,6 +26,7 @@ struct config {
   int mrtdump_file;			/* Configured MRTDump file (sysdep, fd in unix) */
   char *syslog_name;			/* Name used for syslog (NULL -> no syslog) */
   struct rtable_config *master_rtc;	/* Configuration of master routing table */
+  struct iface_patt *router_id_from;	/* Configured list of router ID iface patterns */
 
   u32 router_id;			/* Our Router ID */
   ip_addr listen_bgp_addr;		/* Listening BGP socket should use this address */
@@ -55,28 +55,33 @@ struct config {
 /* Please don't use these variables in protocols. Use proto_config->global instead. */
 extern struct config *config;		/* Currently active configuration */
 extern struct config *new_config;	/* Configuration being parsed */
-extern struct config *old_config;	/* Old configuration when reconfiguration is in progress */
-extern struct config *future_config;	/* New config held here if recon requested during recon */
-
-extern int shutting_down;
-extern bird_clock_t boot_time;
 
 struct config *config_alloc(byte *name);
 int config_parse(struct config *);
 int cli_parse(struct config *);
 void config_free(struct config *);
-int config_commit(struct config *, int type);
-#define RECONFIG_HARD 0
-#define RECONFIG_SOFT 1
+int config_commit(struct config *, int type, int timeout);
+int config_confirm(void);
+int config_undo(void);
+void config_init(void);
 void cf_error(char *msg, ...) NORET;
 void config_add_obstacle(struct config *);
 void config_del_obstacle(struct config *);
 void order_shutdown(void);
 
-#define CONF_DONE 0
-#define CONF_PROGRESS 1
-#define CONF_QUEUED 2
-#define CONF_SHUTDOWN 3
+#define RECONFIG_NONE	0
+#define RECONFIG_HARD	1
+#define RECONFIG_SOFT	2
+#define RECONFIG_UNDO	3
+
+#define CONF_DONE	0
+#define CONF_PROGRESS	1
+#define CONF_QUEUED	2
+#define CONF_UNQUEUED	3
+#define CONF_CONFIRM	4
+#define CONF_SHUTDOWN	-1
+#define CONF_NOTHING	-2
+
 
 /* Pools */
 
@@ -91,7 +96,6 @@ void cfg_copy_list(list *dest, list *src, unsigned node_size);
 /* Lexer */
 
 extern int (*cf_read_hook)(byte *buf, unsigned int max, int fd);
-extern int (*cf_open_hook)(char *filename);
 
 struct symbol {
   struct symbol *next;
@@ -117,12 +121,14 @@ struct symbol {
 #define SYM_VARIABLE 0x100	/* 0x100-0x1ff are variable types */
 
 struct include_file_stack {
-        void	*stack; /* Internal lexer state */
-        unsigned int    conf_lino; /* Current file lineno (at include) */
-        char            conf_fname[BIRD_FNAME_MAX]; /* Current file name */
-        int             conf_fd; /* Current file descriptor */
-        struct include_file_stack *prev;
-        struct include_file_stack *next;
+  void *buffer;				/* Internal lexer state */
+  char *file_name;			/* File name */
+  int fd;				/* File descriptor */
+  int lino;				/* Current line num */
+  int depth;				/* Include depth, 0 = cannot include */
+
+  struct include_file_stack *prev;	/* Previous record in stack */
+  struct include_file_stack *up;	/* Parent (who included this file) */
 };
 
 extern struct include_file_stack *ifs;
