@@ -622,9 +622,6 @@ interpret(struct f_inst *what)
     case T_VOID: runtime( "Can't operate with values of type void" );
     case T_INT: if (v2.val.i == 0) runtime( "Mother told me not to divide by 0" );
       	        res.val.i = v1.val.i / v2.val.i; break;
-    case T_IP: if (v2.type != T_INT)
-                 runtime( "Incompatible types in / operator" );
-               break;
     default: runtime( "Usage of unknown type" );
     }
     break;
@@ -852,12 +849,25 @@ interpret(struct f_inst *what)
     {
       struct rta *rta = (*f_rte)->attrs;
       switch (what->aux) {
-      case T_ENUM:
-	* ((char *) rta + what->a2.i) = v1.val.i;
-	break;
+
       case T_IP:
 	* (ip_addr *) ((char *) rta + what->a2.i) = v1.val.px.ip;
 	break;
+
+      case T_ENUM_SCOPE:
+	rta->scope = v1.val.i;
+	break;
+
+      case T_ENUM_RTD:
+	i = v1.val.i;
+	if ((i != RTD_BLACKHOLE) && (i != RTD_UNREACHABLE) && (i != RTD_PROHIBIT))
+	  runtime( "Destination can be changed only to blackhole, unreachable or prohibit" );
+	rta->dest = i;
+	rta->gw = IPA_NONE;
+	rta->iface = NULL;
+	rta->nexthops = NULL;
+	break;
+
       default:
 	bug( "Unknown type in set of static attribute" );
       }
@@ -943,11 +953,25 @@ interpret(struct f_inst *what)
       l->attrs[0].type = what->aux | EAF_ORIGINATED;
       switch (what->aux & EAF_TYPE_MASK) {
       case EAF_TYPE_INT:
-      case EAF_TYPE_ROUTER_ID:
 	if (v1.type != T_INT)
 	  runtime( "Setting int attribute to non-int value" );
 	l->attrs[0].u.data = v1.val.i;
 	break;
+
+      case EAF_TYPE_ROUTER_ID:
+#ifndef IPV6
+	/* IP->Quad implicit conversion */
+	if (v1.type == T_IP) {
+	  l->attrs[0].u.data = ipa_to_u32(v1.val.px.ip);
+	  break;
+	}
+#endif
+	/* T_INT for backward compatibility */
+	if ((v1.type != T_QUAD) && (v1.type != T_INT))
+	  runtime( "Setting quad attribute to non-quad value" );
+	l->attrs[0].u.data = v1.val.i;
+	break;
+
       case EAF_TYPE_OPAQUE:
 	runtime( "Setting opaque attribute is not allowed" );
 	break;
