@@ -39,6 +39,7 @@ struct protocol {
   char *template;			/* Template for automatic generation of names */
   int name_counter;			/* Counter for automatic name generation */
   int attr_class;			/* Attribute class known to this protocol */
+  int multitable;			/* Protocol handles all announce hooks itself */
   unsigned preference;			/* Default protocol preference */
 
   void (*preconfig)(struct protocol *, struct config *);	/* Just before configuring */
@@ -90,20 +91,26 @@ struct proto_config {
   int class;				/* SYM_PROTO or SYM_TEMPLATE */
   u32 debug, mrtdump;			/* Debugging bitfields, both use D_* constants */
   unsigned preference, disabled;	/* Generic parameters */
+  int in_keep_filtered;			/* Routes rejected in import filter are kept */
   u32 router_id;			/* Protocol specific router ID */
   struct rtable_config *table;		/* Table we're attached to */
   struct filter *in_filter, *out_filter; /* Attached filters */
+  struct proto_limit *rx_limit;		/* Limit for receiving routes from protocol
+					   (relevant when in_keep_filtered is active) */
+  struct proto_limit *in_limit;		/* Limit for importing routes from protocol */
+  struct proto_limit *out_limit;	/* Limit for exporting routes to protocol */
 
   /* Check proto_reconfigure() and proto_copy_config() after changing struct proto_config */
 
   /* Protocol-specific data follow... */
 };
 
-  /* Protocol statistics */
+/* Protocol statistics */
 struct proto_stats {
   /* Import - from protocol to core */
   u32 imp_routes;		/* Number of routes successfully imported to the (adjacent) routing table */
-  u32 pref_routes;		/* Number of routes that are preferred, sum over all routing table */
+  u32 filt_routes;		/* Number of routes rejected in import filter but kept in the routing table */
+  u32 pref_routes;		/* Number of routes that are preferred, sum over all routing tables */
   u32 imp_updates_received;	/* Number of route updates received */
   u32 imp_updates_invalid;	/* Number of route updates rejected as invalid */
   u32 imp_updates_filtered;	/* Number of route updates rejected by filters */
@@ -137,13 +144,16 @@ struct proto {
   u32 debug;				/* Debugging flags */
   u32 mrtdump;				/* MRTDump flags */
   unsigned preference;			/* Default route preference */
-  unsigned accept_ra_types;		/* Which types of route announcements are accepted (RA_OPTIMAL or RA_ANY) */
-  unsigned disabled;			/* Manually disabled */
-  unsigned proto_state;			/* Protocol state machine (see below) */
-  unsigned core_state;			/* Core state machine (see below) */
-  unsigned core_goal;			/* State we want to reach (see below) */
-  unsigned reconfiguring;		/* We're shutting down due to reconfiguration */
-  unsigned refeeding;			/* We are refeeding (valid only if core_state == FS_FEEDING) */
+  byte accept_ra_types;			/* Which types of route announcements are accepted (RA_OPTIMAL or RA_ANY) */
+  byte disabled;			/* Manually disabled */
+  byte proto_state;			/* Protocol state machine (PS_*, see below) */
+  byte core_state;			/* Core state machine (FS_*, see below) */
+  byte core_goal;			/* State we want to reach (FS_*, see below) */
+  byte reconfiguring;			/* We're shutting down due to reconfiguration */
+  byte refeeding;			/* We are refeeding (valid only if core_state == FS_FEEDING) */
+  byte flushing;			/* Protocol is flushed in current flush loop round */
+  byte down_sched;			/* Shutdown is scheduled for later (PDS_*) */
+  byte down_code;			/* Reason for shutdown (PDC_* codes) */
   u32 hash_key;				/* Random key used for hashing of neighbors */
   bird_clock_t last_state_change;	/* Time of last state transition */
   char *last_state_name_announced;	/* Last state name we've announced to the user */
@@ -193,8 +203,7 @@ struct proto {
   void (*rte_remove)(struct network *, struct rte *);
 
   struct rtable *table;			/* Our primary routing table */
-  struct filter *in_filter;		/* Input filter */
-  struct filter *out_filter;		/* Output filter */
+  struct announce_hook *main_ahook;	/* Primary announcement hook */
   struct announce_hook *ahooks;		/* Announcement hooks for this protocol */
 
   struct fib_iterator *feed_iterator;	/* Routing table iterator used during protocol feeding */
@@ -209,6 +218,20 @@ struct proto_spec {
 };
 
 
+#define PDS_DISABLE		1	/* Proto disable scheduled */
+#define PDS_RESTART		2	/* Proto restart scheduled */
+
+#define PDC_CF_REMOVE		0x01	/* Removed in new config */
+#define PDC_CF_DISABLE		0x02	/* Disabled in new config */
+#define PDC_CF_RESTART		0x03	/* Restart due to reconfiguration */
+#define PDC_CMD_DISABLE		0x11	/* Result of disable command */
+#define PDC_CMD_RESTART		0x12	/* Result of restart command */
+#define PDC_CMD_SHUTDOWN	0x13	/* Result of global shutdown */
+#define PDC_RX_LIMIT_HIT	0x21	/* Route receive limit reached */
+#define PDC_IN_LIMIT_HIT	0x22	/* Route import limit reached */
+#define PDC_OUT_LIMIT_HIT	0x23	/* Route export limit reached */
+
+
 void *proto_new(struct proto_config *, unsigned size);
 void *proto_config_new(struct protocol *, unsigned size, int class);
 void proto_copy_config(struct proto_config *dest, struct proto_config *src);
@@ -218,6 +241,10 @@ static inline void
 proto_copy_rest(struct proto_config *dest, struct proto_config *src, unsigned size)
 { memcpy(dest + 1, src + 1, size - sizeof(struct proto_config)); }
 
+
+void proto_show_limit(struct proto_limit *l, const char *dsc);
+void proto_show_basic_info(struct proto *p);
+
 void proto_cmd_show(struct proto *, unsigned int, int);
 void proto_cmd_disable(struct proto *, unsigned int, int);
 void proto_cmd_enable(struct proto *, unsigned int, int);
@@ -344,6 +371,41 @@ void proto_notify_state(struct proto *p, unsigned state);
 
 extern struct proto_config *cf_dev_proto;
 
+
+/*
+ * Protocol limits
+ */
+
+#define PLD_RX		0	/* Receive limit */
+#define PLD_IN		1	/* Import limit */
+#define PLD_OUT		2	/* Export limit */
+#define PLD_MAX		3
+
+#define PLA_WARN	1	/* Issue log warning */
+#define PLA_BLOCK	2	/* Block new routes */
+#define PLA_RESTART	4	/* Force protocol restart */
+#define PLA_DISABLE	5	/* Shutdown and disable protocol */
+
+#define PLS_INITIAL	0	/* Initial limit state after protocol start */
+#define PLS_ACTIVE	1	/* Limit was hit */
+#define PLS_BLOCKED	2	/* Limit is active and blocking new routes */
+
+struct proto_limit {
+  u32 limit;			/* Maximum number of prefixes */
+  byte action;			/* Action to take (PLA_*) */
+  byte state;			/* State of limit (PLS_*) */
+};
+
+void proto_notify_limit(struct announce_hook *ah, struct proto_limit *l, int dir, u32 rt_count);
+
+static inline void
+proto_reset_limit(struct proto_limit *l)
+{
+  if (l)
+    l->state = PLS_INITIAL;
+}
+
+ 
 /*
  *	Route Announcement Hook
  */
@@ -352,18 +414,17 @@ struct announce_hook {
   node n;
   struct rtable *table;
   struct proto *proto;
+  struct filter *in_filter;		/* Input filter */
+  struct filter *out_filter;		/* Output filter */
+  struct proto_limit *rx_limit;		/* Receive limit (for in_keep_filtered) */
+  struct proto_limit *in_limit;		/* Input limit */
+  struct proto_limit *out_limit;	/* Output limit */
+  struct proto_stats *stats;		/* Per-table protocol statistics */
   struct announce_hook *next;		/* Next hook for the same protocol */
+  int in_keep_filtered;			/* Routes rejected in import filter are kept */
 };
 
-struct announce_hook *proto_add_announce_hook(struct proto *, struct rtable *);
-
-/*
- *	Some pipe-specific nest hacks
- */
-
-#ifdef CONFIG_PIPE
-#include "proto/pipe/pipe.h"
-#endif
-
+struct announce_hook *proto_add_announce_hook(struct proto *p, struct rtable *t, struct proto_stats *stats);
+struct announce_hook *proto_find_announce_hook(struct proto *p, struct rtable *t);
 
 #endif
