@@ -112,6 +112,10 @@ ospf_lsa_flooding_allowed(struct ospf_lsa_header *lsa, u32 domain, struct ospf_i
 {    
   u32 scope = LSA_SCOPE(lsa);
 
+  /* Handle inactive vlinks */
+  if (ifa->state == OSPF_IS_DOWN)
+    return 0;
+
   /* 4.5.2 (Case 2) */
   if (unknown_lsa_type(lsa) && !(lsa->type & LSA_UBIT))
     scope = LSA_SCOPE_LINK;
@@ -119,7 +123,7 @@ ospf_lsa_flooding_allowed(struct ospf_lsa_header *lsa, u32 domain, struct ospf_i
   switch (scope)
     {
     case LSA_SCOPE_LINK:
-      return ifa->iface->index == domain;
+      return ifa->iface_id == domain;
 
     case LSA_SCOPE_AREA:
       return ifa->oa->areaid == domain;
@@ -314,7 +318,9 @@ ospf_lsupd_flood(struct proto_ospf *po,
       {
       case OSPF_IT_BCAST:
 	if ((ifa->state == OSPF_IS_BACKUP) || (ifa->state == OSPF_IS_DR))
-	  ospf_send_to(ifa, AllSPFRouters);
+	  ospf_send_to_all(ifa);
+	else if (ifa->cf->real_bcast)
+	  ospf_send_to_bdr(ifa);
 	else
 	  ospf_send_to(ifa, AllDRouters);
 	break;
@@ -327,7 +333,7 @@ ospf_lsupd_flood(struct proto_ospf *po,
 	break;
 
       case OSPF_IT_PTP:
-	ospf_send_to(ifa, AllSPFRouters);
+	ospf_send_to_all(ifa);
 	break;
 
       case OSPF_IT_PTMP:
@@ -500,15 +506,17 @@ ospf_lsupd_receive(struct ospf_packet *ps_i, struct ospf_iface *ifa,
       continue;
     }
 #else /* OSPFv3 */
+    u16 scope = ntoht(lsa->type) & LSA_SCOPE_MASK;
+
     /* 4.5.1 (2) */
-    if ((LSA_SCOPE(lsa) == LSA_SCOPE_AS) && !oa_is_ext(ifa->oa))
+    if ((scope == LSA_SCOPE_AS) && !oa_is_ext(ifa->oa))
     {
       log(L_WARN "Received LSA with AS scope in stub area from %I", n->ip);
       continue;
     }
 
     /* 4.5.1 (3) */
-    if ((LSA_SCOPE(lsa) == LSA_SCOPE_RES))
+    if (scope == LSA_SCOPE_RES)
     {
       log(L_WARN "Received LSA with invalid scope from %I", n->ip);
       continue;
