@@ -8,15 +8,8 @@
 
 #define CONFIG_AUTO_ROUTES
 #define CONFIG_SELF_CONSCIOUS
-#undef CONFIG_MULTIPLE_TABLES
-
-#undef CONFIG_UNIX_IFACE
-#undef CONFIG_UNIX_SET
-#undef CONFIG_UNIX_DONTROUTE
 
 #define CONFIG_SKIP_MC_BIND
-#define CONFIG_ALL_MULTICAST
-#define CONFIG_UNNUM_MULTICAST
 
 /*
 Link: sysdep/unix
